# R. Heeres 2008
# HR@KIT 2017
import qkit
import os
import logging

if qkit.cfg.get('startdir',False):
    os.chdir(qkit.cfg.get('startdir'))


if qkit.cfg.get('startscript',False):
    _scripts = qkit.cfg.get('startscript')
    if type(_scripts) not in (list, tuple):
        _scripts = [_scripts, ]
    for _s in _scripts:
        if os.path.isfile(_s):
            print('Executing (user startscript): %s' % _s)
            execfile(_s)
        else:
            logging.warning('Did not find startscript "%s", skipping'%_s)


if qkit.cfg.get('exitscript',False):
    _scripts = qkit.cfg.get('exitscript')
    if type(_scripts) not in (list, tuple):
        _scripts = [_scripts, ]
    for _s in _scripts:
        if os.path.isfile(_s):
            qkit.flow.register_exit_script(_s)
        else:
            logging.warning('Did not find exitscript "%s", will not be executed', _s)

# Start IPython command logging if requested
if qkit.cfg.get('ipython_logfile',None):
    from IPython import get_ipython
    _ip = get_ipython()
    _ip.IP.logger.logstart(logfname=qkit.cfg.get('ipython_logfile'), logmode='append')



#!/usr/bin/env python

# HP_34401A class, derived from Picowatt_AVS47 class by Hannes Rotzinger.
# Jochen Zimmer, June 2012 -- ALPHA release version 0.1
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


QTLAB = True

if QTLAB:
    from qkit.core.instrument_base import Instrument
else:
    class Instrument(object):
        FLAG_GETSET = 0 
        FLAG_GET_AFTER_SET = 1
        def __init__(self,name,**kwargs):
            pass
        def add_parameter(self,name,**kwargs):
            pass
        def add_function(self,name):
            pass

import visa_prologix as visa
from time import sleep,clock
import sys

class HP_34401A(Instrument):

    def __init__(self, name, address,ip=""):

        Instrument.__init__(self, name, tags=['physical'])

        self._address = address
        self._visainstrument = visa.instrument(self._address,ip=ip)

        self.set_default() #reset

        self.add_function('set_default')
        self.add_function('get_bandwidth')
        self.add_function('set_bandwidth')
        self.add_function('get_auto_zero')
        self.add_function('set_auto_zero')
        self.add_function('get_configuration')
        self.add_function('get_measurement_mode')
        self.add_function('get_range')
        self.add_function('get_resolution')
        self.add_function('set_measurement_mode')
        self.add_function('set_range_resolution')
        self.add_function('get_NPLC')
        self.add_function('set_NPLC')
        self.add_function('get_autorange')
        self.add_function('set_autorange')
        self.add_function('measure')

    def set_default(self):
        com_str = "*RST"
        self._visainstrument.write(com_str)
        sleep(0.5)
        return 0

    def get_bandwidth(self):
        com_str = "DETector:BANDwidth?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(float(res_str))

    #  DETector:BANDwidth {3|20|200|MINimum|MAXimum}
    ## 3 Hz to 300 kHz     7 seconds / reading
    ## 20 Hz to 300 kHz    1 reading / second
    ## 200 Hz to 300 kHz   10 readings / second

    def set_bandwidth(self, bw = 20):
        bw = int(bw)
        if (bw != 3) and (bw != 20) and (bw != 200):
            print >> sys.stderr, "HP_34401A: Input error in set_bandwidth()!"
            return -1
        com_str = "DETector:BANDwidth %i" % bw
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_bandwidth()

    def get_auto_zero(self): # on or off
        com_str = "ZERO:AUTO?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return res_str

# ZERO:AUTO {OFF|ONCE|ON}
# auto-zeroing after each measurement

    def set_auto_zero(self, autoz = "OFF"):
        if (autoz != "OFF") and (autoz != "ONCE") and (autoz != "ON"):
            print >> sys.stderr, "HP_34401A: Input error in set_auto_zero()!"
            return -1
        com_str = "ZERO:AUTO %s" % autoz
        self._visainstrument.write(com_str)
        sleep(0.1)
        return self.get_auto_zero()

    def get_configuration(self):
        com_str = "CONF?"
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        res_str = res_str.split(" ")
        m_mode = res_str[0].split("\"")[1]
        res_str = res_str[1].split(",")
        m_range = float(res_str[0])
        m_resolution = float(res_str[1].split("\"")[0])
        return m_mode, m_range, m_resolution

    def get_measurement_mode(self):
        return self.get_configuration()[0]

    def get_range(self):
        return self.get_configuration()[1]

    def get_resolution(self):
        return self.get_configuration()[2]

    def set_measurement_mode(self, mode = "VOLT:DC"):
        vcf = mode.split(":")[0]
        if (vcf != "VOLT") and (vcf != "CURR") and (vcf != "FREQ"):
            print >> sys.stderr, "HP_34401A: Input error in set_measurement_mode()!"
            return -1
        if (vcf == "FREQ"):
            com_str = "CONF:FREQ"
            self._visainstrument.write(com_str)
            sleep(0.1)
            return self.get_measurement_mode()
        if len(mode.split(":")) < 2:
            print >> sys.stderr, "HP_34401A: Input error in set_measurement_mode()!"
            return -1
        ac_dc = mode.split(":")[1]
        if (ac_dc != "AC") and (ac_dc != "DC"):
            print >> sys.stderr, "HP_34401A: Input error in set_measurement_mode()!"
            return -1
        mode = "%s:%s" % (vcf, ac_dc)
        com_str = "CONF:%s" % mode
        self._visainstrument.write(com_str)
        sleep(0.1)
        return self.get_measurement_mode()

    def set_range_resolution(self,new_range,new_res):
        new_range = float(new_range)
        new_res = float(new_res)
        mode = self.get_measurement_mode()
        com_str = "CONF:%s %f,%f" % (mode, new_range, new_res)
        self._visainstrument.write(com_str)
        sleep(0.1)
        return self.get_range(),self.get_resolution()

    def get_NPLC(self):
        mode = self.get_measurement_mode()
        if (mode != "CURR") and (mode != "VOLT"):
            print >> sys.stderr, "HP_34401A: NPLC not defined"
            return -1
        com_str = "%s:NPLCycles?" % mode
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return float(res_str)

    def set_NPLC(self, plc = 10):  # number of power line cycles, default is 10
        plc = float(plc)
        if (plc != 100) and (plc != 10) and (plc != 1) and (plc != 0.2) and (plc != 0.02):
            print >> sys.stderr, "HP_34401A: NPLC can only be 1, 10 or 100"
            return -1
        mode = self.get_measurement_mode()
        if (mode != "CURR") and (mode != "VOLT"):
            print >> sys.stderr, "HP_34401A: NPLC not defined"
            return -1
        com_str = "%s:NPLCycles %f" % (mode, plc)
        self._visainstrument.write(com_str)
        sleep(0.05)
        return self.get_NPLC()

    def get_autorange(self):
        mode = self.get_measurement_mode()
        com_str = "%s:RANGe:AUTO?" % mode
        self._visainstrument.write(com_str)
        res_str = self._visainstrument.read()
        return int(res_str)

    def set_autorange(self, autoran = 0):
        autoran = int(bool(autoran))
        mode = self.get_measurement_mode()
        com_str = "%s:RANGe:AUTO %i" % (mode, autoran)
        self._visainstrument.write(com_str)
        sleep(0.1)
        return self.get_autorange()

    def measure(self):
        mode = self.get_measurement_mode()
        com_str = "MEAS:%s?" % mode
        self._visainstrument.write(com_str)
        sleep(0.5)
        res_str = self._visainstrument.read()
        return float(res_str)


## testing:

if __name__ == "__main__":

    hp_mm = HP_34401A("HP_34401A", "GPIB::17", ip="192.168.0.100")

    hp_mm._visainstrument.write("*IDN?")
    print hp_mm._visainstrument.read()

#    start = float(clock())
#    for i in range(100):
#        hp_mm.get_configuration()
#    end = float(clock())    
#    print "Getting conf took:", end-start

    print hp_mm.set_measurement_mode("CURR:DC")

    print hp_mm.set_range_resolution(2,0.00001)

    print hp_mm.set_NPLC(10)

    print hp_mm.measure()

    print hp_mm.get_measurement_mode()




# NI_DAQ.py, National Instruments Data Acquisition instrument driver
# Reinier Heeres <reinier@heeres.eu>, 2009
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
# Micha Wildermuth, micha.wildermuth@kit.edu 2019
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# DAQ=qt.instruments.create('DAQ', 'NI_DAQ_IV',id='Dev1')
import qkit
from qkit.core.instrument_base import Instrument
import logging
import nidaq_syncIV as nidaq  # documentation see http://zone.ni.com/reference/en-XX/help/370471AM-01/
import numpy as np
import time


class NI_DAQ_IV(Instrument):
    """
    This is the driver for the National Instruments DAQs such as NI USB-6259 BNC.
    """

    def __init__(self, name, dev, reset=False):
        """
        Initializes VISA communication with the instrument Keithley 2636A.

        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        dev: str
            Device identifier such as 'Dev1'
        reset: bool, optional
            Resets the instrument to default conditions. Default is False.

        Returns
        -------
        None

        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]

        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py

        >>> DAQ = qkit.instruments.create('DAQ', 'NI_DAQ_IV', dev='Dev1', reset=True)
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/mxcprop/func193b/
        self.__name__ = __name__
        # setup
        logging.info(__name__ + ': Initializing instrument NI DAQ')
        Instrument.__init__(self, name, tags=['physical'])
        if dev in nidaq.get_device_names():
            self._dev = dev
        else:
            raise ValueError('{:s}: Cannot find device {!s}'.format(self.__name__, id))
        self._plc = 50.
        self._input_channels = self._get_input_channels()
        self._output_channels = self._get_output_channels()
        self._chan_config = dict(zip(self._input_channels, ['DEFAULT']*len(self._input_channels)))
        self._bias_delay = dict(zip(self._output_channels, np.zeros_like(self._output_channels, dtype=float)))
        self._sense_delay = dict(zip(self._input_channels, np.zeros_like(self._input_channels, dtype=float)))
        self._sample_freq = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)*1e4))
        self._sense_average = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)))
        self._sense_nplc = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)*1e-2))
        self._bias_values = dict(zip(self._output_channels, np.zeros_like(self._output_channels, dtype=float)))
        self._sweep_channels = (0, 0)
        self.bias_data = None
        self.rate = None
        # reset
        if reset:
            self.reset()
        else:
            self.get_all()

    def write(self, data, channel, freq=1e4, minv=-10.0, maxv=10.0, timeout=10.0):
        """
        Writes analog output values <data> to output channel <channel> with sample frequency <freq>.

        Parameters
        ----------
        data: int/float/numpy.array
            Voltage data to write in Volts.
        channel: int
            Number of output channel of interest.
        freq: float
            Sample frequency in Hertz.
        minv: float
            Minimum voltage in Volts.
        maxv: float
            Maximum voltage in Volts.
        timeout: float
            Time to wait for completion in seconds.

        Returns
        -------
        written: int
            Number of values written
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcreatetask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcreateaovoltagechan/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxwriteanalogscalarf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcfgsampclktiming/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxwriteanalogf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstarttask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstoptask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcleartask/
        return nidaq.write(devchan='{:s}/ao{:d}'.format(self._dev, channel),
                           data=data,
                           freq=freq,
                           minv=minv,
                           maxv=maxv,
                           timeout=timeout)

    def read(self, channel, samples=1, freq=1e4, minv=-10.0, maxv=10.0, timeout=10.0):
        """
        Reads analog input values <data> of input channel <channel> with sample frequency <freq>.

        Parameters
        ----------
        channel: int
            Number of input channel of interest.
        samples: int
            Number of samples to read up to max_samples.
        freq: float
            Sample frequency in Hertz.
        minv: float
            Minimum voltage in Volts.
        maxv: float
            Maximum voltage in Volts.
        timeout: float
            Time to wait for completion in seconds.

        Returns
        -------
        data: float/numpy.array/None
            Read voltage data in Volts on success or None on error.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcreatetask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471W-01/daqmxcfunc/daqmxcreateaivoltagechan/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcfgsampclktiming/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstarttask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxreadanalogf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxreadanalogscalarf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstoptask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcleartask/
        return nidaq.read(devchan='{:s}/ai{:d}'.format(self._dev, channel),
                          samples=samples,
                          freq=freq,
                          minv=minv,
                          maxv=maxv,
                          timeout=timeout,
                          config=self._chan_config[channel])

    def _get_input_channels(self):
        """
        Gets all physical input channels of the device.

        Parameters
        ----------
        None

        Returns
        -------
        channels: list(str)
            Physical input channels.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/mxcprop/func231e/
        return np.array(list(map(lambda x: x.split('ai')[-1], nidaq.get_physical_input_channels(self._dev))), dtype=int)

    def _get_output_channels(self):
        """
        Gets all physical output channels of the device.

        Parameters
        ----------
        None

        Returns
        -------
        channels: list(str)
            Physical output channels.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/mxcprop/func231f/
        return np.array(list(map(lambda x: x.split('ao')[-1], nidaq.get_physical_output_channels(self._dev))), dtype=int)

    def _get_counter_channels(self):
        """
        Gets all physical counter channels of the device.

        Parameters
        ----------
        None

        Returns
        -------
        channels: list(str)
            Physical counter channels.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/mxcprop/func2324/
        return np.array(list(map(lambda x: x.split('ctr')[-1], nidaq.get_physical_counter_channels(self._dev))), dtype=int)

    def set_channel_config(self, val, channel):
        """
        Sets the channel configuration of channel <channel> to the mode <val>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).
        val: str
            Analog input channel configuration. Possible values are
              * 'DEFAULT',
              * 'RSE' (Referenced single-ended mode),
              * 'NRSE' (Non-referenced single-ended mode),
              * 'Diff' (Differential mode),
              * 'PseudoDiff' (Pseudodifferential mode)

        Returns
        -------
        None
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471W-01/daqmxcfunc/daqmxcreateaivoltagechan/
        self._chan_config[channel] = val.upper()

    def get_channel_config(self, channel):
        """
        Gets the channel configuration of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).

        Returns
        -------
        val: str
            Analog input channel configuration. Possible values are
              * 'DEFAULT',
              * 'RSE' (Referenced single-ended mode),
              * 'NRSE' (Non-referenced single-ended mode),
              * 'Diff' (Differential mode),
              * 'PseudoDiff' (Pseudodifferential mode)
        """
        return self._chan_config[channel]

    ####################################################################################################################
    ### functions needed for transport.py and virtual_tunnel_electronics.py                                          ###
    ####################################################################################################################

    def get_measurement_mode(self, channel=0):
        """
        This is a dummy function and is always 1 (4-wire). This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        return 1  # 4-wire

    def get_bias_mode(self, channel=0):
        """
        This is a dummy function and is always 1 (voltage). This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        return 1  # voltage bias

    def get_sense_mode(self, channel=0):
        """
        This is a dummy function and is always 1 (voltage). This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        return 1  # voltage bias

    def get_bias_range(self, channel=0):
        """
        This is a dummy function and is always -1 ('auto'). This DAQ provides only the range -10V .. +10V.
        """
        return -1  # 10V

    def get_sense_range(self, channel=0):
        """
        This is a dummy function and is always -1 ('auto'). This DAQ provides only the range -10V .. +10V.
        """
        return -1  # 10V

    def get_bias_delay(self, channel=0):
        """
        This is a dummy function and is always 0. This DAQ samples continuously and provides no delay.
        """
        return 0

    def get_sense_delay(self, channel=0):
        """
        This is a dummy function and is always 0. This DAQ samples continuously and provides no delay.
        """
        return 0

    def set_sense_average(self, val, channel=0):
        """
        Sets sense average of channel <channel> to <val>.

        Parameters
        ----------
        val: int
            Number of measured readings that are required to yield one filtered measurement.
        channel: int
            Number of channel of interest. Default is 0 (ai0).

        Returns
        -------
        None
        """
        self._sense_average[channel] = int(val)

    def get_sense_average(self, channel=0):
        """
        Gets sense average status <status>, value <val> and mode <mode> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).

        Returns
        -------
        val: int
            Number of measured readings that are required to yield one filtered measurement.
        """
        return self._sense_average[channel]

    def set_plc(self, val):
        """
        Sets power line cycle (PLC) to <val>.

        Parameters
        ----------
        val: int
            Power line frequency setting used for NPLC calculations. 50 | 60.

        Returns
        -------
        None
        """
        self._plc = val

    def get_plc(self):
        """
        Gets power line cycle (PLC) <plc>.

        Parameters
        ----------
        None

        Returns
        -------
        plc: int
            Power line frequency setting used for NPLC calculations.
        """
        return self._plc

    def set_sense_nplc(self, val, sample_freq=None, channel=0):
        """
        Sets sense nplc (number of power line cycle) of channel <channel> with the <val>-fold of one power line cycle. Note that the sample frequency needs to be >= plc/nplc.
        The integral of sampled data is calculated by menas of numpy.trapz, which approximates the integral by trapezoidal rule (https://en.wikipedia.org/wiki/Trapezoidal_rule)

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).
        val: float
            Integration aperture for measurements.
        sample_freq: float
            Sample frequency. Default is None (previous settings).

        Returns
        -------
        None
        """
        if sample_freq:
            self._sample_freq[channel] = sample_freq
        self._sense_nplc[channel] = val
        if not int(self._sample_freq[channel] / self._plc * self._sense_nplc[channel]):
            logging.warning('Measurement time will increase, since samples will be set to 1. Samples ( = sample_freq / plc * nplc) must be >= 1. Check the sampling frequency and nplc settings.')

    def get_sense_nplc(self, channel=0):
        """
        Gets sense nplc (number of power line cycle) <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).

        Returns
        -------
        val: float
            Integration aperture for measurements.
        """
        return self._sense_nplc[channel]

    def set_sample_freq(self, val, channel=0):
        """
        Sets sample frequency of input channel <channel> to <val>.

        Parameters
        ----------
        channel: int
            Number of input channel of interest. Default is 0 (ai0).
        val: float
            Sample frequency. Default is None (previous settings).

        Returns
        -------
        None
        """
        self._sample_freq[channel] = val

    def get_sample_freq(self, channel=0):
        """
        Gets sample frequency of input channel <channel>.

        Parameters
        ----------
        channel: int
            Number of input channel of interest. Default is 0 (ai0).

        Returns
        -------
        val: float
            Sample frequency. Default is None (previous settings).
        """
        return self._sample_freq[channel]

    def set_bias_value(self, val, channel=0):
        """
        Sets bias value of channel <channel> to value <val>.

        Parameters
        ----------
        val: float
            Bias value.
        channel: int
            Number of channel of interest. Default is 0 (ao0).

        Returns
        -------
        None
        """
        self._bias_values[channel] = val
        nidaq.write(devchan='{:s}/ao{:d}'.format(self._dev, channel), data=val)

    def get_bias_value(self, channel=0):
        """
        Gets bias value <val> of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ao0).

        Returns
        -------
        val: float
            Bias value.
        """
        # TODO: read back bias value (https://knowledge.ni.com/KnowledgeArticleDetails?id=kA00Z000000kGDISA2&l=de-DE)
        # TODO: than use bias_delay
        return self._bias_values[channel]

    def get_sense_value(self, sample_freq=None, channel=0):
        """
        Gets sense value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0 (ai0).
        
        Returns
        -------
        val: float
            Sense value.
        """
        if sample_freq:
            self._sample_freq[channel] = sample_freq
        samples = int(self._sample_freq[channel] / self._plc * self._sense_nplc[channel])
        if not samples:
            samples = 1
        data = list(map(lambda i: nidaq.read(devchan='{:s}/ai{:d}'.format(self._dev, channel),
                                             samples=samples,
                                             freq=self._sample_freq[channel],
                                             config=self._chan_config[channel]),
                        range(int(self._sense_average[channel]))))
        return np.mean(np.trapz(data, dx=1./(samples-1)))

    def set_status(self, status, channel=0):
        """
        Sets output status of channel <channel> to <status>.
        This is a dummy function. The output cannot be turned off, so that False equals 0V.

        Parameters
        ----------
        status: int
            Output status.
        channel: int
            Number of channel of interest. Default is 0 (ao0).

        Returns
        -------
        None
        """
        if not status:
            self.set_bias_value(val=0, channel=channel)

    def set_sweep_channels(self, *channels):
        """
        Sets channels <channels> that are used during sweeps.

        Parameters
        ----------
        channels: int or tuple(int)
            Number of channel of usage. The 1st argument is the bias channel and the 2nd argument the sense channel.

        Returns
        -------
        None
        """
        self._sweep_channels = channels

    def get_sweep_channels(self):
        """
        Gets channels <channels> that are used during sweeps.

        Parameters
        ----------
        None

        Returns
        -------
        channels: int or tuple(int)
            Number of channel of usage. The 1st argument is the bias channel and the 2nd argument the sense channel.
        """
        return self._sweep_channels

    def get_sweep_bias(self):
        """
        This is a dummy function and is always 1 (voltage). This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        return 1  # voltage

    def set_sweep_mode(self, *args, **kwargs):
        """
        This is a dummy function and just passes. This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        pass

    def get_sweep_mode(self):
        """
        This is a dummy function and is always 0 (VV-mode). This DAQ can only source and measure voltages and currents by means of tunnel electronics.
        """
        return 0  # (VV-mode)

    def set_sweep_parameters(self, sweep):
        """
        Sets sweep parameters <sweep>.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)

        Returns
        -------
        None
        """
        start = float(sweep[0])
        stop = float(sweep[1])
        step = float(sweep[2])
        self.nop = int(round(abs((stop-start)/step)+1))
        self.bias_data = np.linspace(start, np.sign(stop-start)*(np.floor(np.abs(np.round(float(stop-start)/step)))*step)+start, self.nop)  # stop is rounded down to multiples of step
        self.rate = self._plc/self._sense_nplc[self.get_sweep_channels()[1]]
        self.set_bias_value(start)
        time.sleep(1./self.rate)

    def get_tracedata(self):
        """
        Starts bias sweep with already set parameters and gets trace data of bias <bias_values> and sense <sense_values> in the set sweep mode.
        Note that this function sweeps the bias permanently, reads synchronized with a 10fold sampling rate and averages this 10 samples
        The sweep rate is calculated from plc and nplc, while averages are ignored.

        Parameters
        ----------
        None

        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/mxcprop/func2986/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcreatetask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471W-01/daqmxcfunc/daqmxcreateaivoltagechan/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcfgsampclktiming/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcreateaovoltagechan/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcfgdigedgestarttrig/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxwriteanalogf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstarttask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxwaituntiltaskdone/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxreadanalogf64/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxstoptask/
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxcleartask/
        output_channel, input_channel = [''.join(i).strip() for i in zip(*np.array([[self._dev] * 2,
                                                                                    ['/ao', '/ai'],
                                                                                    self.get_sweep_channels()]))]
        sense_data = nidaq.sync_write_read(O_devchan=output_channel,  # device / bias channel
                                           I_devchan=input_channel,  # device / sense channel
                                           waveform=self.bias_data,  # bias data
                                           rate=self.rate,  # sweep rate
                                           timeout=self.nop/self.rate*1.1,  # 1.1 fold of the calculated measurement time, just to be on the safe side
                                           config=self._chan_config[self.get_sweep_channels()[1]])  # sense channel configuration
        return self.bias_data, sense_data

    def take_IV(self, sweep):
        """
        Takes IV curve with sweep parameters <sweep>.

        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)

        Returns
        -------
        bias_values: numpy.array(float)
            Bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        self.set_sweep_parameters(sweep=sweep)
        return self.get_tracedata()

    def reset(self):
        """
        Resets the instrument to default conditions.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        # Corresponding command: http://zone.ni.com/reference/en-XX/help/370471AM-01/daqmxcfunc/daqmxresetdevice/
        nidaq.reset_device(self._dev)

    def set_defaults(self):
        """
        Sets default settings.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.reset()
        self._plc = 50.
        self._chan_config = dict(zip(self._input_channels, ['DEFAULT']*len(self._input_channels)))
        self._bias_delay = dict(zip(self._output_channels, np.zeros_like(self._output_channels, dtype=float)))
        self._sense_delay = dict(zip(self._input_channels, np.zeros_like(self._input_channels, dtype=float)))
        self._sample_freq = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)*1e4))
        self._sense_average = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)))
        self._sense_nplc = dict(zip(self._input_channels, np.ones_like(self._input_channels, dtype=float)*1e-2))
        self._bias_values = dict(zip(self._output_channels, np.zeros_like(self._output_channels, dtype=float)))
        self._bias_values = None
        self.rate = None
        self._sweep_channels = (0, 0)

    def get_all(self, channel=0):
        """
        Prints all settings of channel <channel>.

        Parameters
        ----------
        channel: int
            Number of channel of interest. Default is 0.

        Returns
        -------
        None
        """
        logging.debug('{:s}: Get all'.format(__name__))
        print('output channels  = {!s}'.format(self._output_channels))
        print('input channels   = {!s}'.format(self._input_channels))
        print('channel config   = {!s}'.format(self._chan_config[channel]))
        print('sample frequency = {:g}'.format(self._sample_freq[channel]))
        print('sense average      = {:f}'.format(self.get_sense_average(channel=channel)))
        print('plc                = {:g}Hz'.format(self.get_plc()))
        print('sense nplc         = {:g}'.format(self.get_sense_nplc(channel=channel)))
        print('bias value         = {:g}V'.format(self.get_bias_value(channel=channel)))
        print('sense value        = {:g}V'.format(self.get_sense_value(channel=channel)))
        return

    def get_parameters(self):
        """
        Gets a parameter list <parlist> of measurement specific setting parameters.
        Needed for .set-file in 'write_additional_files', if qt parameters are not used.

        Parameters
        ----------
        None

        Returns
        -------
        parlist: dict
            Parameter names as keys, corresponding channels of interest as values.
        """
        parlist = {'channel_config': self._input_channels,
                   #'bias_delay': self._output_channels,
                   #'sense_delay': self._input_channels,
                   'sample_freq': self._input_channels,
                   'sense_average': self._input_channels,
                   'sense_nplc': self._input_channels,
                   'bias_value': self._output_channels,
                  }
        return parlist

    def get(self, param, **kwargs):
        """
        Gets the current parameter <param> by evaluation 'get_'+<param> and corresponding channel if needed
        In combination with <self.get_parameters> above.

        Parameters
        ----------
        param: str
            Parameter of interest.
        channels: array_likes
            Number of channels of interest. Can be a list for channel specific parameters or None for channel independent (global) parameters.

        Returns
        -------
        parlist: dict
            Parameter names as keys, values of corresponding channels as values.
        """
        channels = kwargs.get('channels')
        if channels is [None]:
            return tuple(eval('map(lambda channel, self=self: self.get_{:s}(), [None])'.format(param)))
            #return eval('self.get_{:s}()'.format(param))
        else:
            return tuple(eval('map(lambda channel, self=self: self.get_{:s}(channel=channel), [{:s}])'.format(param, ', '.join(map(str, channels)))))
            #return tuple([eval('self.get_{:s}(channel={!s})'.format(param, channel)) for channel in channels])



import qkit
#qkit.cfg["run_id"] = "Magpie_2020_05_19"
#qkit.cfg["user"] = "Martin"
qkit.start()

import time
import math
import numpy as np
import matplotlib.pyplot as plt
from ipywidgets import interact, interactive, fixed, interact_manual

from scipy.signal import savgol_filter
from scipy.optimize import curve_fit
from scipy import fftpack

from scipy.optimize import curve_fit

from qkit.measure.timedomain.quantum_machine import qm_qubit_config as qc
from qkit.measure.timedomain.quantum_machine import basic_qubit_experiments as bqe
from qkit.measure.timedomain.quantum_machine import quantum_jump_monitoring as qjm
from qkit.measure.timedomain.quantum_machine import pulses
from qkit.analysis import iq_cloud_analysis

qm_config = qc.QMQubitConfig()
qm_config.wait_sig = 156
qm_config.smearing_sig = 48
qm_config.wait_ref = 62
qm_config.smearing_ref = 0

qm_config.readout_pulse_len = 128
qm_config.set_readout_wf(pulses.const, (0.49,))
qm_config.set_pi_pulse_wf(pulses.gauss, (0.3630, 0.0, 10.0))

qm = bqe.BasicQubitExperiments(qm_config)
qmjm = qjm.QuantumJumpMonitoring(qm_config)

iqcloud = iq_cloud_analysis.IQCloudAnalysis()
iqcloud.set_up_gaussian_mixture_model(2)

##### launch experiment ####

ref_pos = np.array([[0.63315114, 0.75654685], [-0.49696001, 0.78736931]])
var = np.array([0.04187039, 0.04051063])
dist_avg = 0.14078334743527843

ts = time.time()

i = 0

qmjm.dirname = "t_rep_2p0us_sigma_1p0_run_{}".format(i)
qmjm.quantum_jump_hist_stab_avg(25, 10, ref_pos * dist_avg, (np.sqrt(var) * dist_avg) * np.array([1.0, 1.0]),
                            t_stab=242, t_measure=230, avg=10,
                            stab_state=True, start_state=False, plot=True, save=False)

#t_stab = 242, t_measure = 480,
#242, t_measure = 230,

print("Ready")






# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from time import sleep
import types
import logging
try:
    from qkit.core.instrument_base import Instrument
except:
    # debug mode
    logging.getLogger().setLevel(logging.INFO)
    logging.warning(__name__ + ': unable to load instrument module. entering standalone mode.')
    class Instrument:
        FLAG_SET = 0x01
        FLAG_GET = 0x02
        def __init__(self, name, tags):
            pass
        def add_parameter(self, name, type, flags, channels = (), minval = None, maxval = None, units = None, format = None, tags = None):
            pass
        def add_function(self, func):
            pass
from pyftdi.pyftdi.ftdi import Ftdi
from pyftdi.pyftdi.usbtools import UsbTools 

'''
    Tunnelelektronik DAC w/ FTDI USB interface
'''
class Tunnel_DAC(Instrument):

    def __init__(self, name, serial = None, channel = 'A0', numdacs=3, delay = 1e-3):
        '''
            discover and initialize Tunnel_DAC hardware
            
            Input:
                serial - serial number of the FTDI converter
                channel - 2 character channel id the DAC is connected to;
                    the first byte identifies the channel (A..D for current devices)
                    the second byte identifies the bit within that channel (0..7)
                numdacs - number of DACs daisy-chained on that line
                delay - communications delay assumed between PC and the USB converter
        '''
        logging.info(__name__+ ': Initializing instrument Tunnel_DAC')
        Instrument.__init__(self, name, tags=['physical'])
        
        self._conn = Ftdi()
        # VIDs and PIDs of converters used
        vps = [
            (0x0403, 0x6011), # FTDI UM4232H 4ch
            (0x0403, 0x6014)  # FTDI UM232H 1ch
        ]
        # explicitly clear device cache of UsbTools
        #UsbTools.USBDEVICES = []
        # find all devices and obtain serial numbers
        devs = self._conn.find_all(vps)
        # filter list by serial number if provided
        if(serial != None):
            devs = [dev for dev in devs if dev[2] == serial]
        if(len(devs) == 0):
            logging.error(__name__ + ': failed to find matching FTDI devices.')
        elif(len(devs) > 1):
            logging.error(__name__ + ': more than one converter found and no serial number given.')
            logging.info(__name__ + ': available devices are: %s.'%str([dev[2] for dev in devs]))
        vid, pid, self._serial, channels, description = devs[0]
        # parse channel string
        if(len(channel) != 2):
            logging.error(__name__ + ': channel identifier must be a string of length 2. ex. A0, D5.')
        self._channel = 1 + ord(channel[0]) - ord('A')
        self._bit = ord(channel[1]) - ord('0')
        if((self._channel < 1) or (self._channel > channels)):
            logging.error(__name__ + ': channel %c is not supported by this device.'%(chr(ord('A')+self._channel-1)))
        if((self._bit < 0) or (self._bit > 7)):
            logging.error(__name__ + ': subchannel must be between 0 and 7, not %d.'%self._bit)

        # open device
        self._conn.open(vid, pid, interface = self._channel, serial = self._serial)
        logging.info(__name__ + ': using converter with serial #%s'%self._serial)
        self._conn.set_bitmode(0xFF, Ftdi.BITMODE_BITBANG)
        # 80k generates bit durations of 12.5us, 80 is magic :(
        # magic?: 4 from incorrect BITBANG handling of pyftdi, 2.5 from 120MHz instead of 48MHz clock of H devices
        # original matlab code uses 19kS/s
        self._conn.set_baudrate(19000/80)

        # house keeping        
        self._numdacs = numdacs
        self._sleeptime = (10. + 16.*self._numdacs)*12.5e-6 + delay # 1st term from hardware parameters, 2nd term from USB  
        self._minval = -5000.
        self._maxval = 5000.
        self._resolution = 16 # DAC resolution in bits
        self._voltages = [0.]*numdacs
        
        self.add_parameter('voltage', type=float, flags=Instrument.FLAG_SET, channels=(1, self._numdacs),
                           minval = self._minval, maxval = self._maxval, units='mV', format = '%.02f') # tags=['sweep']
        self.add_function('set_voltages')
        self.add_function('commit')


    def _encode(self, data, channel = 0, bits_per_item = 16, big_endian = True):
        '''
            convert binary data into line symbols
            
            the tunnel electronic DAC logic box triggers on rising 
            signal edges and samples data after 18us. we use a line
            code with three bits of duration 12us, where a logical 1
            is encoded as B"110" and a logical 0 is encoded as B"100".
            
            the line data is returned as a byte string with three bytes/symbol.
            
            Input:
                data - a vector of data entities (usually a string or list of integers)
                channel - a number in [0, 7] specifying the output bit on the USB-to-UART chip
                bits_per_item - number of bits to extract from each element of the data vector
        '''
        # build line code for the requested channel
        line_1 = chr(1<<channel)
        line_0 = chr(0)
        line_code = [''.join([line_0, line_1, line_1]), ''.join([line_0, line_1, line_0])]
        # do actual encoding
        result = []
        result.append(10*line_0)
        for item in data:
            for bit in (range(bits_per_item-1, -1, -1) if big_endian else range(0, bits_per_item)):
                result.append(line_code[1 if(item & (1<<bit)) else 0])
        result.append(10*line_0)
        return ''.join(result)


    def commit(self):
        '''
            send updated parameter values to the physical DACs via USB
        '''
        # normalize, scale, clip voltages
        voltages = [-1+2*(x-self._minval)/(self._maxval-self._minval) for x in self._voltages]
        voltages = [max(-2**(self._resolution-1), min(2**(self._resolution-1)-1, int(2**(self._resolution-1)*x))) for x in voltages]
        # encode and send
        data = self._encode(reversed(voltages), self._bit, self._resolution)
        self._conn.write_data(data)
        # wait for the FTDI fifo to clock the data to the DACs
        sleep(self._sleeptime)


    def do_set_voltage(self, value, channel):
        '''
            immediately update voltage on channel ch
            parameter checking is done by qtlab
        '''
        self._voltages[channel-1] = value
        self.commit()


    def set_voltages(self, valuedict):
        '''
            update voltages on several channels simultaneously
            todo: update instrument panel display
        '''
        for channel, value in valuedict.iteritems():
            # bounds checking & clipping
            if((channel < 1) or (channel > self._numdacs)):
                logging.error(__name__ + ': channel %d out of range.'%channel)
                continue
            value = float(value)
            if((value < self._minval) or (value >= self._maxval)):
                logging.error(__name__ + ': value %f out of range. clipping.'%value)
                value = max(self._minval, min(self._maxval, value)) # does not handle maxval correctly
            self._voltages[channel-1] = value
        self.commit()



# modified and adapted by JB@KIT 04/2015, 09/2015
# modified by MK@KIT 09/2018
# time domain measurement class

import numpy as np
import logging
import threading

import qkit
from qkit.gui.notebook.Progress_Bar import Progress_Bar
from qkit.storage import store as hdf
from qkit.gui.plot import plot as qviewkit
import qkit.measure.write_additional_files as waf
from qkit.measure.timedomain.initialize import InitializeTimeDomain as iniTD
from qkit.measure.measurement_class import Measurement


class Measure_td(object):
    '''
    useage:

    m = Measure_td()

    m.set_x_parameters(arange(-0.05,0.05,0.01),'flux coil current (mA)',coil.set_current)
    m.set_y_parameters(arange(4e9,7e9,10e6),'excitation frequency (Hz)',mw_src1.set_frequency)

    m.measure_XX()
    Generally, we want to use
    ReadoutTrace = True -> if we want to record the readout pulse or
    AWGTrace = True -> if we have a N different time steps in the AWG (Not used, this is done via the mode variable now.)

    ToDO (S1, 09/2017):
        - Include LogFunctions
        - Include 2D_AWG with pre-averaging
        - Check multi-tone readout
    '''
    
    def __init__(self, sample, readout=None):
        self.sample = sample
        if readout is None:
            self.readout = sample.readout
        else:
            self.readout = readout
        
        self.comment = None
        self.mode = None
        self.x_set_obj = None
        self.y_set_obj = None
        self.z_set_obj = None
        
        self.dirname = None
        self.data_dir = None
        self.plotSuffix = ''
        self.hold = False
        
        self.show_progress_bar = True
        
        self.ReadoutTrace = False
        
        self.open_qviewkit = True
        self.create_averaged_data = False
        
        self.qviewkit_singleInstance = True
        self._qvk_process = False
        self._plot_comment = ''
        self.multiplex_attribute = "readout_pulse_frequency"
        self.multiplex_unit = "Hz"
        self.init = iniTD(sample)

        self._measurement_object = Measurement()
        self._measurement_object.measurement_type = 'timedomain'
        self._measurement_object.sample = self.sample
        
    
    def set_x_parameters(self, x_vec, x_coordname, x_set_obj, x_unit=None):
        self.x_vec = x_vec
        self.x_coordname = x_coordname
        self.x_set_obj = x_set_obj
        if x_unit is None:
            logging.warning(__name__ + ': Unit of the x-axis is not set.')
            self.x_unit = ''
        else:
            self.x_unit = x_unit
    
    def set_y_parameters(self, y_vec, y_coordname, y_set_obj, y_unit=None):
        self.y_vec = y_vec
        self.y_coordname = y_coordname
        self.y_set_obj = y_set_obj
        if y_unit is None:
            logging.warning(__name__ + ': Unit of the y-axis is not set.')
            self.y_unit = ''
        else:
            self.y_unit = y_unit
    
    def set_z_parameters(self, z_vec, z_coordname, z_set_obj, z_unit=None):
        self.z_vec = z_vec
        self.z_coordname = z_coordname
        self.z_set_obj = z_set_obj
        if z_unit is None:
            logging.warning(__name__ + ': Unit of the z-axis is not set.')
            self.z_unit = ''
        else:
            self.z_unit = z_unit
    
    def measure_1D(self):
        if self.x_set_obj is None:
            raise ValueError('x-axes parameters not properly set')
        
        self.mode = 1  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG, 4:3D_AWG
        self._prepare_measurement_file()
        
        if self.show_progress_bar: p = Progress_Bar(len(self.x_vec), name=self.dirname)
        try:
            # measurement loop
            for x in self.x_vec:
                self.x_set_obj(x)
                qkit.flow.sleep()
                self._append_data()
                if self.show_progress_bar: p.iterate()
        finally:
            self._end_measurement()
    
    def measure_2D(self):
        if self.x_set_obj is None or self.y_set_obj is None:
            raise ValueError('Axes parameters not properly set')
        if self.ReadoutTrace:
            raise ValueError('ReadoutTrace is currently not supported for 2D measurements')
        
        self.mode = 2  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG, 4:3D_AWG
        self._prepare_measurement_file()
        
        if self.show_progress_bar: p = Progress_Bar(len(self.x_vec) * len(self.y_vec), name=self.dirname)
        try:
            # measurement loop
            for x in self.x_vec:
                self.x_set_obj(x)
                for y in self.y_vec:
                    qkit.flow.sleep()
                    self.y_set_obj(y)
                    qkit.flow.sleep()
                    self._append_data()
                    if self.show_progress_bar: p.iterate()
                for i in range(self.ndev):
                    self._hdf_amp[i].next_matrix()
                    self._hdf_pha[i].next_matrix()
        finally:
            self._end_measurement()
    
    def measure_1D_AWG(self, iterations=100):
        '''
        use AWG sequence for x_vec, averaging over iterations
        '''
        self.y_vec = range(iterations)
        self.y_coordname = 'iteration'
        self.y_set_obj = lambda y: True
        self.y_unit = ''
        self.create_averaged_data = True
        self.avg_complex_sum = np.zeros_like(self.x_vec)
        try:
            return self.measure_2D_AWG(iterations=1)
        finally:
            self.create_averaged_data = False  # This is ALWAYS done after the return! Looks strange and it really is, but it works.
    
    def measure_2D_AWG(self, iterations=1):
        '''
        x_vec is sequence in AWG
        '''
        
        if self.y_set_obj is None:
            raise ValueError('y-axes parameters not properly set')
        
        qkit.flow.sleep()  # if stop button was pressed by now, abort without creating data files
        
        if iterations > 1:
            self.z_vec = range(iterations)
            self.z_coordname = 'iteration'
            self.z_set_obj = lambda z: True
            self.z_unit = ''
            
            self.measure_3D_AWG()
            
            # For 3D measurements with iterations, averages are only created at the end to get a consistent averaging base.
            hdf_file = hdf.Data(self._hdf.get_filepath())
            for j in range(self.ndev):
                amp = np.array(hdf_file["/entry/data0/amplitude_%i" % j])
                pha = np.array(hdf_file["/entry/data0/phase_%i" % j])
                amp_avg = sum(amp[i] for i in range(iterations)) / iterations
                pha_avg = sum(pha[i] for i in range(iterations)) / iterations
                hdf_amp_avg = hdf_file.add_value_matrix('amplitude_avg_%i' % i, x=self._hdf_y, y=self._hdf_x, unit='a.u.')
                hdf_pha_avg = hdf_file.add_value_matrix('phase_avg_%i' % i, x=self._hdf_y, y=self._hdf_x, unit='rad')
                
                for i in range(len(self.y_vec)):
                    hdf_amp_avg.append(amp_avg[i], pointwise=True)
                    hdf_pha_avg.append(pha_avg[i], pointwise=True)
            hdf_file.close_file()
        
        else:
            self.mode = 3  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG, 4:3D_AWG
            self._prepare_measurement_file()
            # if self.ndev > 1: raise ValueError('Multiplexed readout is currently not supported for 2D measurements')
            if self.show_progress_bar:
                p = Progress_Bar(len(self.y_vec), name=self.dirname)
            try:
                # measurement loop
                for it in range(len(self.y_vec)):
                    qkit.flow.sleep()  # better done during measurement (waiting for trigger)
                    self.y_set_obj(self.y_vec[it])
                    self._append_data(iteration=it)
                    if self.show_progress_bar: p.iterate()
            finally:
                self._end_measurement()
    
    def measure_3D_AWG(self):
        '''
        x_vec is sequence in AWG
        '''
        
        if self.z_set_obj is None or self.y_set_obj is None:
            raise ValueError('x-axes parameters not properly set')
        if self.ReadoutTrace:
            raise ValueError('ReadoutTrace is currently not supported for 3D_AWG measurements')
        
        self.mode = 4  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG, 4:3D_AWG
        self._prepare_measurement_file()
        
        if self.show_progress_bar: p = Progress_Bar(len(self.y_vec) * len(self.z_vec), name=self.dirname)
        try:
            # measurement loop
            for z in self.z_vec:
                self.z_set_obj(z)
                for y in self.y_vec:
                    qkit.flow.sleep()
                    self.y_set_obj(y)
                    qkit.flow.sleep()
                    self._append_data()
                    if self.show_progress_bar: p.iterate()
                for i in range(self.ndev):
                    self._hdf_amp[i].next_matrix()
                    self._hdf_pha[i].next_matrix()
        finally:
            self._end_measurement()
    
    def measure_1D_ddc_time_trace(self):
        """
        measures the time evolution of your transmission / reflection
        in your network by performing a digital down conversion
        :return: None
        """
        time_end = float(self.sample.mspec.get_samples()) / self.sample.mspec.get_samplerate()
        time_array = np.linspace(0, time_end, self.sample.mspec.get_samples())
        self.set_x_parameters(time_array, 'time', True, 'sec')
        self.mode = 1  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG
        self._prepare_measurement_file()
        try:
            qkit.flow.sleep()
            self._append_data(ddc=True)
        finally:
            self._end_measurement()
    
    def measure_2D_ddc_time_trace(self):
        """
        Performs a digital down conversion for exactly one value in your awg sequence. But you can sweep other
        parameters, such as mw-power or so.
        :return:
        """
        if self.y_set_obj is None:
            raise ValueError('y-axes parameters not properly set')
        time_end = float(self.sample.mspec.get_samples()) / self.sample.mspec.get_samplerate()
        time_array = np.linspace(0, time_end, self.sample.mspec.get_samples())
        self.set_x_parameters(time_array, 'time', True, 'sec')
        
        self.mode = 2  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG
        self._prepare_measurement_file()
        
        if self.show_progress_bar:
            p = Progress_Bar(len(self.y_vec), name=self.dirname)
        try:
            for y in self.y_vec:
                qkit.flow.sleep()
                self.y_set_obj(y)
                qkit.flow.sleep()
                self._append_data(ddc=True)
                if self.show_progress_bar: p.iterate()
        finally:
            self._end_measurement()
    
    def measure_1D_awg_ddc_timetrace(self):
        """
        Performs a digital down conversion of your readout trace for every value in your y-vector,
        whereas y_vec should be awg sequence. This function is useful for magnon cavity experiments.
        x-vec is automatically set by acquisition window.
        (Also, all data are there at once)
        :return:
        """
        if self.y_vec is None:
            raise ValueError('y-axes parameters not properly set')
        time_end = float(self.sample.mspec.get_samples()) / self.sample.mspec.get_samplerate()
        time_array = np.linspace(0, time_end, self.sample.mspec.get_samples())
        self.set_x_parameters(time_array, 'time', True, 'sec')
        
        self.mode = 2  # 1: 1D, 2: 2D, 3:1D_AWG/2D_AWG
        self._prepare_measurement_file()
        
        try:
            qkit.flow.sleep()
            self._append_data(ddc=True)
        finally:
            self._end_measurement()
    
    def measure_1D_awg_1D_ddc_timetrace(self):
        """
        Performs a digital down conversion of your readout trace for every value in your y-vector
        while also sweeping another parameter. y_vec should be awg sequence and z_vec/obj the sweeping parameter.
        x-vec is automatically set by acquisition window.
        :return:
        """
        if (self.y_vec is None) or (self.z_set_obj is None):
            raise ValueError('Axes parameters not properly set')
        time_end = float(self.sample.mspec.get_samples()) / self.sample.mspec.get_samplerate()
        time_array = np.linspace(0, time_end, self.sample.mspec.get_samples())
        self.set_x_parameters(time_array, 'time', True, 'sec')
        self.mode = 4
        self._prepare_measurement_file()
        try:
            for z in self.z_vec:
                self.z_set_obj(z)
                qkit.flow.sleep()
                self._append_data(ddc=True)
                for i in range(self.ndev):
                    self._hdf_amp[i].next_matrix()
                    self._hdf_pha[i].next_matrix()
                if self.ReadoutTrace:
                    self._hdf_I.next_matrix()
                    self._hdf_Q.next_matrix()
        finally:
            self._end_measurement()
    
    def _prepare_measurement_file(self):
        qkit.flow.start()
        if self.dirname is None:
            self.dirname = self.x_coordname
        
        self.ndev = len(self.readout.get_tone_freq())  # returns array of readout freqs (=1 for non-multiplexed readout)
        
        self._hdf = hdf.Data(name=self.dirname, mode='a')
        self._hdf_x = self._hdf.add_coordinate(self.x_coordname, unit=self.x_unit)
        self._hdf_x.add(self.x_vec)

        self._measurement_object.uuid = self._hdf._uuid
        self._measurement_object.hdf_relpath = self._hdf._relpath
        self._measurement_object.instruments = qkit.instruments.get_instrument_names()

        self._measurement_object.save()
        self._mo = self._hdf.add_textlist('measurement')
        self._mo.append(self._measurement_object.get_JSON())

        self._settings = self._hdf.add_textlist('settings')
        settings = waf.get_instrument_settings(self._hdf.get_filepath())
        self._settings.append(settings)
        
        self._log = waf.open_log_file(self._hdf.get_filepath())
        
        self._hdf_readout_frequencies = self._hdf.add_coordinate(self.multiplex_attribute, unit=self.multiplex_unit)
        self._hdf_readout_frequencies.add(self.readout.get_tone_freq())
        
        if self.ReadoutTrace:
            self._hdf_TimeTraceAxis = self._hdf.add_coordinate('recorded timepoint', unit='s')
            self._hdf_TimeTraceAxis.add(np.arange(self.sample.mspec.get_samples()) / self.readout.get_adc_clock())
        
        if self.mode == 1:  # 1D
            self._hdf_amp = []
            self._hdf_pha = []
            for i in range(self.ndev):
                self._hdf_amp.append(self._hdf.add_value_vector('amplitude_%i' % i, x=self._hdf_x, unit='a.u.'))
                self._hdf_pha.append(self._hdf.add_value_vector('phase_%i' % i, x=self._hdf_x, unit='rad'))
            if self.ReadoutTrace:
                self._hdf_I = self._hdf.add_value_matrix('I_TimeTrace', x=self._hdf_x, y=self._hdf_TimeTraceAxis,
                                                         unit='V', save_timestamp=False)
                self._hdf_Q = self._hdf.add_value_matrix('Q_TimeTrace', x=self._hdf_x, y=self._hdf_TimeTraceAxis,
                                                         unit='V', save_timestamp=False)
        
        elif self.mode == 2:  # 2D
            self._hdf_y = self._hdf.add_coordinate(self.y_coordname, unit=self.y_unit)
            self._hdf_y.add(self.y_vec)
            self._hdf_amp = []
            self._hdf_pha = []
            for i in range(self.ndev):
                self._hdf_amp.append(self._hdf.add_value_matrix('amplitude_%i' % i, x=self._hdf_x, y=self._hdf_y, unit='a.u.'))
                self._hdf_pha.append(self._hdf.add_value_matrix('phase_%i' % i, x=self._hdf_x, y=self._hdf_y, unit='rad'))
            if self.ReadoutTrace:
                # TODO: One dimension missing here?
                self._hdf_I = self._hdf.add_value_matrix('I_TimeTrace', x=self._hdf_y, y=self._hdf_TimeTraceAxis,
                                                         unit='V', save_timestamp=False)
                self._hdf_Q = self._hdf.add_value_matrix('Q_TimeTrace', x=self._hdf_y, y=self._hdf_TimeTraceAxis,
                                                         unit='V', save_timestamp=False)
        
        elif self.mode == 3:  # 1D_AWG/2D_AWG
            self._hdf_y = self._hdf.add_coordinate(self.y_coordname, unit=self.y_unit)
            self._hdf_y.add(self.y_vec)
            self._hdf_amp = []
            self._hdf_pha = []
            for i in range(self.ndev):
                self._hdf_amp.append(self._hdf.add_value_matrix('amplitude_%i' % i,
                                                                x=self._hdf_y, y=self._hdf_x, unit='a.u.'))
                self._hdf_pha.append(self._hdf.add_value_matrix('phase_%i' % i,
                                                                x=self._hdf_y, y=self._hdf_x, unit='rad'))
            if self.ReadoutTrace:
                self._hdf_I = self._hdf.add_value_box('I_TimeTrace', x=self._hdf_y, y=self._hdf_x,
                                                      z=self._hdf_TimeTraceAxis, unit='V', save_timestamp=False)
                self._hdf_Q = self._hdf.add_value_box('Q_TimeTrace', x=self._hdf_y, y=self._hdf_x,
                                                      z=self._hdf_TimeTraceAxis, unit='V', save_timestamp=False)
        
        elif self.mode == 4:  # 3D_AWG
            self._hdf_y = self._hdf.add_coordinate(self.y_coordname, unit=self.y_unit)
            self._hdf_y.add(self.y_vec)
            self._hdf_z = self._hdf.add_coordinate(self.z_coordname, unit=self.z_unit)
            self._hdf_z.add(self.z_vec)
            self._hdf_amp = []
            self._hdf_pha = []
            for i in range(self.ndev):
                self._hdf_amp.append(self._hdf.add_value_box('amplitude_%i' % i, x=self._hdf_z, y=self._hdf_y,
                                                             z=self._hdf_x, unit='a.u.'))
                self._hdf_pha.append(self._hdf.add_value_box('phase_%i' % i, x=self._hdf_z, y=self._hdf_y,
                                                             z=self._hdf_x, unit='rad'))
            if self.ReadoutTrace:
                self._hdf_I = self._hdf.add_value_box('I_TimeTrace', x=self._hdf_z, y=self._hdf_y,
                                                      z=self._hdf_TimeTraceAxis, unit='V', save_timestamp=False)
                self._hdf_Q = self._hdf.add_value_box('Q_TimeTrace', x=self._hdf_y, y=self._hdf_y,
                                                      z=self._hdf_TimeTraceAxis, unit='V', save_timestamp=False)
        
        if self.create_averaged_data:
            self._hdf_amp_avg = []
            self._hdf_pha_avg = []
            for i in range(self.ndev):
                self._hdf_amp_avg.append(self._hdf.add_value_vector('amplitude_avg_%i' % i, x=self._hdf_x, unit='a.u.'))
                self._hdf_pha_avg.append(self._hdf.add_value_vector('phase_avg_%i' % i, x=self._hdf_x, unit='rad'))
        
        if self.comment:
            self._hdf.add_comment(self.comment)
        self._hdf.hf.hf.attrs['default_ds'] = ['data0/amplitude_%i' % i for i in range(min(5,self.ndev))] +\
                                              ['data0/phase_%i' % i for i in range(min(5,self.ndev))]
        if self.qviewkit_singleInstance and self.open_qviewkit and self._qvk_process:
            self._qvk_process.terminate()  # terminate an old qviewkit instance
        if self.open_qviewkit:
            self._qvk_process = qviewkit.plot(self._hdf.get_filepath(),
                                              datasets=['amplitude_%i' % i for i in range(min(5,self.ndev))] + ['phase_%i' % i for i in range(min(5,self.ndev))]
                                              )
    
        try:
            self.readout.start()
        except AttributeError:
            pass
    
    def _append_data(self, iteration=0, ddc=None):
        if self.ReadoutTrace:
            ampliData, phaseData, Is, Qs = self.readout.readout(timeTrace=True, ddc=ddc)
        else:
            ampliData, phaseData = self.readout.readout(timeTrace=False, ddc=ddc)
        
        if len(ampliData.shape) < 3:  # "normal" measurements
            for i in range(self.ndev):
                self._hdf_amp[i].append(np.atleast_1d(ampliData.T[i]), pointwise=True)
                self._hdf_pha[i].append(np.atleast_1d(phaseData.T[i]), pointwise=True)
            if self.ReadoutTrace:
                if self.mode < 3:  # mode 2 not yet fully supported but working for DDC timetrace experiments
                    self._hdf_I.append(Is, pointwise=True)
                    self._hdf_Q.append(Qs, pointwise=True)
                elif self.mode == 3:  # mode 4 not supported for 3D_awg yet
                    for ix in range(len(self.x_vec)):
                        self._hdf_I.append(Is[:, ix], pointwise=True)
                        self._hdf_Q.append(Qs[:, ix], pointwise=True)
                    self._hdf_I.next_matrix()
                    self._hdf_Q.next_matrix()
        
        else:  # for AWG DDC ReadoutTrace, all data are there at once
            for i in range(self.ndev):
                for j in range(ampliData.T.shape[2]):
                    self._hdf_amp[i].append(np.atleast_1d(ampliData.T[i, :, j]), pointwise=True)
                    self._hdf_pha[i].append(np.atleast_1d(phaseData.T[i, :, j]), pointwise=True)
                    if self.ReadoutTrace:
                        self._hdf_I.append(Is[:, j], pointwise=True)
                        self._hdf_Q.append(Qs[:, j], pointwise=True)
        
        if self.create_averaged_data:
            if iteration == 0:
                self.avg_complex_sum = ampliData * np.exp(1j * phaseData)
                for i in range(self.ndev):
                    self._hdf_amp_avg[i].append(np.atleast_1d(ampliData.T[i]), pointwise=True)
                    self._hdf_pha_avg[i].append(np.atleast_1d(phaseData.T[i]), pointwise=True)
            else:
                self.avg_complex_sum += ampliData * np.exp(1j * phaseData)
                amp_avg = np.abs(self.avg_complex_sum / (iteration + 1))
                pha_avg = np.angle(self.avg_complex_sum / (iteration + 1))
                for i in range(self.ndev):
                    self._hdf_amp_avg[i].ds.write_direct(np.ascontiguousarray(np.atleast_1d(amp_avg.T[i])))
                    self._hdf_pha_avg[i].ds.write_direct(np.ascontiguousarray(np.atleast_1d(pha_avg.T[i])))
                    self._hdf_pha_avg[i].ds.attrs['iteration'] = iteration + 1
                    self._hdf_amp_avg[i].ds.attrs['iteration'] = iteration + 1
                self._hdf.flush()
    
    def _end_measurement(self):
        try:
            self.readout.cleanup()
        except AttributeError:
            pass
        t = threading.Thread(target=qviewkit.save_plots, args=[self._hdf.get_filepath(), self._plot_comment])
        t.start()
        self._hdf.close_file()
        waf.close_log_file(self._log)
        qkit.flow.end()
    
    def set_plot_comment(self, comment):
        '''
        Small comment to add at the end of plot pics for more information i.e. good for wiki entries.
        '''
        self._plot_comment = comment


#example1 = qt.instruments.create('example1', 'example', address='GPIB::1', reset=True)
#dsgen = qt.instruments.create('dsgen', 'dummy_signal_generator')
#pos = qt.instruments.create('pos', 'dummy_positioner')
#combined = qt.instruments.create('combined', 'virtual_composite')
#combined.add_variable_scaled('magnet', example1, 'chA_output', 0.02, -0.13, units='mT')

#combined.add_variable_combined('waveoffset', [{
#    'instrument': dmm1,
#    'parameter': 'ch2_output',
#    'scale': 1,
#    'offset': 0}, {
#    'instrument': dsgen,
#    'parameter': 'wave',
#    'scale': 0.5,
#    'offset': 0
#    }], format='%.04f')



# -*- coding: utf-8 -*-

# qplexkit/server.py is a zmq server that has to run on the Raspberry Pi to
# controll the homemade qplexkit used for DC multiplexing.
# Micha Wildermuth, micha.wildermuth@kit.edu 2023
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import zmq
from zmq.auth.thread import ThreadAuthenticator
import logging
from qkit.services.qplexkit.qplexkit import qplexkit
from qkit.config.services import cfg
import re

##################################################
### setup zeroMQ-server for qkit communication ###
##################################################
''' run qplexkit '''
qpk = qplexkit()

''' dictionary of '''
msg2cmd = {'set_switch_time': qpk.set_switch_time,
           'get_switch_time': qpk.get_switch_time,
           'set_experiment': qpk.set_experiment,
           'get_experiment': qpk.get_experiment,
           'set_current_divider': qpk.set_current_divider,
           'get_current_divider': qpk.get_current_divider,
           'set_amplifier': qpk.set_amplifier,
           'get_amplifier': qpk.get_amplifier,
           'set_current_source_status': qpk.set_current_source_status,
           'get_current_source_status': qpk.get_current_source_status,
           'set_relay': qpk.set_relay,
           'get_relay': qpk.get_relay,
           'get_relays': qpk.get_relays,
           'get_ccr': qpk.get_ccr,
           'read_ccr': qpk.read_ccr,
           'reset': qpk.reset,
           'get_attr': lambda attr, *args, **kwargs: getattr(qpk, attr),
           'ping': lambda *args, **kwargs: 'pong',
           }

''' run zmq-server '''
if __name__ == "__main__":
    context = zmq.Context()
    auth = ThreadAuthenticator(context, log=logging.getLogger())
    auth.start()
    auth.allow('127.0.0.0/8')
    for ip in cfg['qplexkit']['allowed_ip_addresses'].split(','):
        auth.allow(re.compile(r'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})').search(ip)[0])
    socket = context.socket(zmq.REP)
    socket.zap_domain = b'global'
    socket.bind(f'''tcp://*:{cfg['qplexkit']['server_port']}''')

    while True:
        try:
            fun, args, kwargs = socket.recv_json()
            socket.send_json(msg2cmd[fun](*args, **kwargs))
        except Exception as e:
            socket.send_json((e.__class__.__name__, *e.args))


###################################################
### setup zeroRPC-server for qkit communication ###
###################################################
"""
import zerorpc
_address = cfg['qplexkit_address']
s = zerorpc.Server(qplexkit())
s.bind(_address)
s.run()
"""



# nidaq.py, python wrapper for NIDAQ DLL
# Reinier Heeres <reinier@heeres.eu>, 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import ctypes
import numpy as np
import logging
import time

nidaq = ctypes.windll.nicaiu

int32 = ctypes.c_long
int64 = ctypes.c_longlong
uInt32 = ctypes.c_ulong
uInt64 = ctypes.c_ulonglong
float64 = ctypes.c_double
TaskHandle = uInt64

DAQmx_Val_Cfg_Default = int32(-1)
DAQmx_Val_RSE = 10083
DAQmx_Val_NRSE = 10078
DAQmx_Val_Diff = 10106
DAQmx_Val_PseudoDiff = 12529
_config_map = {'DEFAULT': DAQmx_Val_Cfg_Default,
               'RSE': DAQmx_Val_RSE,
               'NRSE': DAQmx_Val_NRSE,
               'DIFF': DAQmx_Val_Diff,
               'PSEUDODIFF': DAQmx_Val_PseudoDiff,
               }

DAQmx_Val_Volts = 10348
DAQmx_Val_Rising = 10280
DAQmx_Val_FiniteSamps = 10178
DAQmx_Val_GroupByChannel = 0
DAQmx_Val_GroupByScanNumber = 1

DAQmx_Val_CountUp = 10128
DAQmx_Val_CountDown = 10124
DAQmx_Val_ExtControlled = 10326


def CHK(err):
    """
    Error checking routine raises an error, if it gets an error code.

    Parameters
    ----------
    err: int
        Error number.

    Returns
    -------
    None
    """
    if err < 0:
        buf_size = 500
        buf = ctypes.create_string_buffer(b'\000' * buf_size)
        nidaq.DAQmxGetErrorString(err, ctypes.byref(buf), buf_size)
        raise RuntimeError('Nidaq call failed with error %d: %s' % (err, repr(buf.value)))

def buf_to_list(buf):
    """
    Decodes a buffer to a python list

    Parameters
    ----------
    buf: str
        Buffer to be decoded and split to a list.

    Returns
    -------
    list: list
        List generated from buffer.
    """
    return buf.raw.decode('ascii').rstrip('\x00').split(', ')

def get_device_names():
    """
    Gets a list of available NI DAQ devices.

    Parameters
    ----------
    None

    Returns
    -------
    names: list(str)
        list of available NI DAQ devices
    """
    bufsize = 1024
    buf = ctypes.create_string_buffer(b'\000' * bufsize)
    nidaq.DAQmxGetSysDevNames(ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def reset_device(dev):
    """
    Resets device.

    Parameters
    ----------
    dev: str
        Device name to be reset.

    Returns
    -------
    None
    """
    nidaq.DAQmxResetDevice(dev.encode('ascii'))

def get_physical_input_channels(dev):
    """
    Gets all physical input channels of a device.

    Parameters
    ----------
    dev: str
        Device name of interest.

    Returns
    -------
    channels: list(str)
        Physical input channels.
    """
    bufsize = 1024
    buf = ctypes.create_string_buffer(b'\000' * bufsize)
    nidaq.DAQmxGetDevAIPhysicalChans(dev.encode('ascii'), ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def get_physical_output_channels(dev):
    """
    Gets all physical output channels of a device.

    Parameters
    ----------
    dev: str
        Device name of interest.

    Returns
    -------
    channels: list(str)
        Physical output channels.
    """
    bufsize = 1024
    buf = ctypes.create_string_buffer(b'\000' * bufsize)
    nidaq.DAQmxGetDevAOPhysicalChans(dev.encode('ascii'), ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def get_physical_counter_channels(dev):
    """
    Gets all physical counter channels of a device.

    Parameters
    ----------
    dev: str
        Device name of interest.

    Returns
    -------
    channels: list(str)
        Physical counter channels.
    """
    bufsize = 1024
    buf = ctypes.create_string_buffer(b'\000' * bufsize)
    nidaq.DAQmxGetDevCIPhysicalChans(dev.encode('ascii'), ctypes.byref(buf), bufsize)
    return buf_to_list(buf)

def write(devchan, data, freq=1e4, minv=-10.0, maxv=10.0, timeout=10.0):
    """
    Writes analog output values <data> to channel <devchan> with sample frequency <freq>.

    Parameters
    ----------
    devchan: str
        Device/output_channel specifier, such as 'Dev1/ao0'.
    data: int/float/np.array
        Voltage data to write in Volts.
    freq: float
        Sample frequency in Hertz.
    minv: float
        Minimum voltage in Volts.
    maxv: float
        Maximum voltage in Volts.
    timeout: float
        Time to wait for completion in seconds.

    Returns
    -------
    written: int
        Number of values written
    """
    if isinstance(data, float) or isinstance(data, int):
        data = np.array([data], dtype=np.float64)
    elif isinstance(data, np.ndarray) and data.dtype is not np.float64:
        data = np.array(data, dtype=np.float64)
    elif data.size > 0:
        data = np.array(data, dtype=np.float64)
    samples = data.size
    AO_TaskHandle = TaskHandle(0)
    written = int32()
    try:
        CHK(nidaq.DAQmxCreateTask(b"", ctypes.byref(AO_TaskHandle)))
        CHK(nidaq.DAQmxCreateAOVoltageChan(AO_TaskHandle,
                                           devchan.encode('ascii'),
                                           b"",
                                           float64(minv),
                                           float64(maxv),
                                           DAQmx_Val_Volts,
                                           None))
        if data.size == 1:
            CHK(nidaq.DAQmxWriteAnalogScalarF64(AO_TaskHandle,
                                                1,
                                                float64(timeout),
                                                float64(data),
                                                None))
            written = int32(1)
        else:
            CHK(nidaq.DAQmxCfgSampClkTiming(AO_TaskHandle,
                                            b"",
                                            float64(freq),
                                            DAQmx_Val_Rising,
                                            DAQmx_Val_FiniteSamps,
                                            uInt64(samples)))
            CHK(nidaq.DAQmxWriteAnalogF64(AO_TaskHandle,
                                          samples,
                                          0,
                                          float64(timeout),
                                          DAQmx_Val_GroupByChannel,
                                          data.ctypes.data,
                                          ctypes.byref(written),
                                          None))
            CHK(nidaq.DAQmxStartTask(AO_TaskHandle))
    except Exception as e:
        logging.error('NI DAQ call failed (correct channel configuration selected?): %s', str(e))
    finally:
        if AO_TaskHandle.value != 0:
            nidaq.DAQmxStopTask(AO_TaskHandle)
            nidaq.DAQmxClearTask(AO_TaskHandle)
    return written.value

def read(devchan, samples=1, freq=1e4, minv=-10.0, maxv=10.0, timeout=10.0, config=DAQmx_Val_Cfg_Default):
    """
    Reads analog input values <data> of channel <devchan> with sample frequency <freq>.

    Parameters
    ----------
    devchan: str
        Device/input_channel specifier, such as 'Dev1/ai0'.
    samples: int
        Number of samples to read up to max_samples.
    freq: float
        Sample frequency in Hertz.
    minv: float
        Minimum voltage in Volts.
    maxv: float
        Maximum voltage in Volts.
    timeout: float
        Time to wait for completion in seconds.
    config: str/int
        Channel configuration.

    Returns
    -------
    data: float/np.array/None
        Read voltage data in Volts on success or None on error.
    """
    if isinstance(config, str):
        if config in _config_map:
            config = _config_map[config]
        else:
            return None
    if not type(config) in (int32, int64, uInt32, uInt64):
        return None
    if samples == 1:
        retsamples = 1
        samples = 2
    else:
        retsamples = samples
    data = np.zeros(samples, dtype=np.float64)
    AI_TaskHandle = TaskHandle(0)
    read = int32()
    try:
        CHK(nidaq.DAQmxCreateTask(b"", ctypes.byref(AI_TaskHandle)))
        CHK(nidaq.DAQmxCreateAIVoltageChan(AI_TaskHandle,
                                           devchan.encode('ascii'),
                                           b"",
                                           config,
                                           float64(minv),
                                           float64(maxv),
                                           DAQmx_Val_Volts,
                                           None))
        if retsamples > 1:
            CHK(nidaq.DAQmxCfgSampClkTiming(AI_TaskHandle,
                                            b"",
                                            float64(freq),
                                            DAQmx_Val_Rising,
                                            DAQmx_Val_FiniteSamps,
                                            uInt64(samples)))
            CHK(nidaq.DAQmxStartTask(AI_TaskHandle))
            CHK(nidaq.DAQmxReadAnalogF64(AI_TaskHandle,
                                         samples,
                                         float64(timeout),
                                         DAQmx_Val_GroupByChannel,
                                         ctypes.c_int64(data.ctypes.data),
                                         samples,
                                         ctypes.byref(read),
                                         None))
        else:
            CHK(nidaq.DAQmxReadAnalogScalarF64(AI_TaskHandle,
                                               float64(timeout),
                                               ctypes.c_int64(data.ctypes.data),
                                               None))
            read = int32(1)
    except Exception as e:
        print(str(e))
        logging.error('NI DAQ call failed: %s', str(e))
    finally:
        if AI_TaskHandle.value != 0:
            nidaq.DAQmxStopTask(AI_TaskHandle)
            nidaq.DAQmxClearTask(AI_TaskHandle)
    if read.value > 0:
        if retsamples == 1:
            return data[0]
        else:
            return data[:read.value]
    else:
        return None

def read_counter(devchan, samples=1, freq=1.0, timeout=1.0, src=""):
    """
    Reads counter values <data> of channel <devchan> with sample frequency <freq>.

    Parameters
    ----------
    devchan: str
        Device/input_channel specifier, such as 'Dev1/ctr0'.
    samples: int
        Number of samples to read up to max_samples.
    freq: float
        Sample frequency in Hertz.
    timeout: float
        Time to wait for completion in seconds.
    src: str
        Specified source pin.

    Returns
    -------
    data: float/np.array/None
        Read counter data on success or None on error.
    """
    Crt_TaskHandle = TaskHandle(0)
    try:
        CHK(nidaq.DAQmxCreateTask(b"", ctypes.byref(Crt_TaskHandle)))
        initial_count = int32(0)
        CHK(nidaq.DAQmxCreateCICountEdgesChan(Crt_TaskHandle,
                                              devchan.encode('ascii'),
                                              b"",
                                              DAQmx_Val_Rising,
                                              initial_count,
                                              DAQmx_Val_CountUp))
        if src is not None and src != "":
            CHK(nidaq.DAQmxSetCICountEdgesTerm(Crt_TaskHandle,
                                               devchan.encode('ascii'),
                                               src))
        nread = int32()
        data = np.zeros(samples, dtype=np.float64)
        if samples > 1:
            CHK(nidaq.DAQmxCfgSampClkTiming(Crt_TaskHandle,
                                            b"",
                                            float64(freq),
                                            DAQmx_Val_Rising,
                                            DAQmx_Val_FiniteSamps,
                                            uInt64(samples)))
            CHK(nidaq.DAQmxStartTask(Crt_TaskHandle))
            CHK(nidaq.DAQmxReadAnalogF64(Crt_TaskHandle,
                                         int32(samples),
                                         float64(timeout),
                                         DAQmx_Val_GroupByChannel,
                                         ctypes.c_int64(data.ctypes.data),
                                         samples,
                                         ctypes.byref(read),
                                         None))
        else:
            CHK(nidaq.DAQmxStartTask(Crt_TaskHandle))
            time.sleep(1.0 / freq)
            nread = int32(0)
            CHK(nidaq.DAQmxReadCounterF64(Crt_TaskHandle,
                                          int32(samples),
                                          float64(timeout),
                                          ctypes.c_int64(data.ctypes.data),
                                          int32(samples),
                                          ctypes.byref(nread),
                                          None))
            nread = int32(1)
    except Exception as e:
        logging.error('NI DAQ call failed: %s', str(e))
    finally:
        if Crt_TaskHandle.value != 0:
            nidaq.DAQmxStopTask(Crt_TaskHandle)
            nidaq.DAQmxClearTask(Crt_TaskHandle)
    if nread.value == 1:
        return int(data[0])
    else:
        return data

def sync_write_read(O_devchan, I_devchan, waveform, rate=1e3, minv=-10.0, maxv=10.0, timeout=10.0, config=DAQmx_Val_Cfg_Default):
    """
    Writes values <waveform> to an output channel <O_devchan> and reads synchronized an input channel <I_devchan> with rate <rate>.

    Parameters
    ----------
    O_devchan: strs
        Device/output_channel specifier, such as 'Dev1/ao0'.
    I_devchan: str
        Device/input_channel specifier, such as 'Dev1/ai0'.
    waveform: list/np.array
        Voltage data to write to O_devchan in Volts.
    rate: float
        Write/read rate in Hertz.
    minv: float
        Minimum voltage in Volts.
    maxv: float
        Maximum voltage in Volts.
    timeout: float
        Time to wait for completion in seconds.
    config: str/int
        Channel configuration.

    Returns
    -------
    data: np.array
        Sense data.
    """
    if type(config) is str:
        if config in _config_map:
            config = _config_map[config]
        else:
            return None
    if not type(config) in (int, int32, uInt32, uInt64):
        return None
    debug = False
    oversamples = 10
    sampswritten = int32()
    sampsread = int32()
    # cast data to np.array
    waveform = np.array(waveform, dtype=np.float64)
    samples = waveform.size
    # Due to Nyquist theorem: sampling rate and samples have to be doubled for reading
    # align fix: After the trigger we read (samples+1) samples and throw away the first
    # (trigger sample) later
    AI_rate = oversamples * rate
    AI_samples = oversamples * samples
    readbuf = np.zeros(AI_samples + 1, dtype=np.float64)
    # Check if the StartTrigger option is available
    DAQmx_Val_Bit_TriggerUsageTypes_Start = 8
    triggerAvailable = ctypes.c_int32()
    CHK(nidaq.DAQmxGetDevAITrigUsage(I_devchan.encode('ascii'), ctypes.byref(triggerAvailable)))
    if not ((int(triggerAvailable.value) & DAQmx_Val_Bit_TriggerUsageTypes_Start) == DAQmx_Val_Bit_TriggerUsageTypes_Start):
        raise RuntimeError('Error: Input trigger for synchronization is not available on this device')
    # create task handles for input and output
    AI_TaskHandle = TaskHandle(0)
    AO_TaskHandle = TaskHandle(1)
    # prepare Analog Input
    CHK(nidaq.DAQmxCreateTask(str(AI_TaskHandle.value),
                              ctypes.byref(AI_TaskHandle)))
    CHK(nidaq.DAQmxCreateAIVoltageChan(AI_TaskHandle,
                                       I_devchan.encode('ascii'),
                                       b"",
                                       config,
                                       float64(minv),
                                       float64(maxv),
                                       DAQmx_Val_Volts,
                                       None))
    CHK(nidaq.DAQmxCfgSampClkTiming(AI_TaskHandle,
                                    b"",
                                    float64(AI_rate),
                                    DAQmx_Val_Rising,
                                    DAQmx_Val_FiniteSamps,
                                    uInt64(AI_samples + 1)))
    if debug:
        print('Debug: Analog Input AI_TaskHandle prepared')
    # prepare Analog Output
    CHK(nidaq.DAQmxCreateTask(str(AO_TaskHandle.value),
                              ctypes.byref(AO_TaskHandle)))
    CHK(nidaq.DAQmxCreateAOVoltageChan(AO_TaskHandle,
                                       O_devchan.encode('ascii'),
                                       b"",
                                       float64(minv),
                                       float64(maxv),
                                       DAQmx_Val_Volts,
                                       None))
    CHK(nidaq.DAQmxCfgSampClkTiming(AO_TaskHandle,
                                    b"",
                                    float64(rate),
                                    DAQmx_Val_Rising,
                                    DAQmx_Val_FiniteSamps,
                                    uInt64(samples)))
    if debug:
        print('Debug: Analog Output AO_TaskHandle prepared')
    # register internal trigger, AI_TaskHandle starts with AO_TaskHandle
    # /Dev1/ao/StartTrigger is the reference name to the output start trigger
    triggerName = '/'+O_devchan.split('/')[0]+'/ao/StartTrigger'
    CHK(nidaq.DAQmxCfgDigEdgeStartTrig(AI_TaskHandle,
                                       triggerName.encode('ascii'),
                                       DAQmx_Val_Rising))
    CHK(nidaq.DAQmxWriteAnalogF64(AO_TaskHandle,
                                  samples,
                                  0,
                                  float64(timeout),
                                  DAQmx_Val_GroupByChannel,
                                  ctypes.c_int64(waveform.ctypes.data),
                                  ctypes.byref(sampswritten),
                                  None))
    if debug:
        print('Debug: read_write() Analog-Output buffer written:', waveform, sampswritten)
    # start the tasks, sampling of (samples +1) at AI_TaskHandle is started when AO_TaskHandle started
    CHK(nidaq.DAQmxStartTask(AI_TaskHandle))
    CHK(nidaq.DAQmxStartTask(AO_TaskHandle))
    CHK(nidaq.DAQmxWaitUntilTaskDone(AI_TaskHandle,
                                     float64(timeout)))
    CHK(nidaq.DAQmxReadAnalogF64(AI_TaskHandle,
                                 AI_samples + 1,
                                 float64(timeout),
                                 DAQmx_Val_GroupByChannel,
                                 ctypes.c_int64(readbuf.ctypes.data),
                                 AI_samples + 1,
                                 ctypes.byref(sampsread),
                                 None))
    if debug:
        print('Debug: read_write() Analog-Input buffer read:', readbuf, readbuf.size)
    # stop the whole thing, fixme: use try...except !
    if AO_TaskHandle.value != 0:
        nidaq.DAQmxStopTask(AO_TaskHandle)
        nidaq.DAQmxClearTask(AO_TaskHandle)
    if AI_TaskHandle.value != 0:
        nidaq.DAQmxStopTask(AI_TaskHandle)
        nidaq.DAQmxClearTask(AI_TaskHandle)
    # again align fix: throw away the first sample and average over the read samples
    # to get the shape than put in
    returnbuf = np.zeros(samples)
    aveis = np.arange(oversamples)
    if debug:
        print(len(readbuf), len(returnbuf), aveis)
    for i in np.arange(samples):
        for ii in aveis:
            returnbuf[i] += readbuf[oversamples * i + ii]
    return returnbuf / float(oversamples)


if __name__ == '__main__':
    # some tests
    buf = np.linspace(0, 5, 10)
    print(sync_write_read('Dev1/ai0', 'Dev1/ao0', buf, rate=100000))




# -*- coding: utf-8 -*-
# Micha Wildermuth, micha.wildermuth@kit.edu 2020

from qkit.core.instrument_base import Instrument
import numpy as np

def get_IVC_JJ(x, Ic, Rn, SNR):
    sign = np.sign(x[-1] - x[0])
    return Rn * x * np.heaviside(np.abs(x) - Ic, int(sign > 0)) \
           + (np.heaviside(x, int(sign > 0)) - np.heaviside(x + sign * Ic, 0)) * Ic * Rn \
           + Ic * Rn / SNR * np.random.rand(x.size)


class IVD_dummy(Instrument):
    '''
    This is a driver for a dummy IV-Device as used for transport measurements with qkit.measure.transport.transport.transport.

    Usage:
        Initialize with
        <name> = qkit.instruments.create('<name>', 'IVD_dummy')
    '''
    def __init__(self, name):
        self.__name__ = __name__
        Instrument.__init__(self, name, tags=['virtual'])
        self.func = get_IVC_JJ
        self.args = ()
        self.kwargs = {'Ic': 1e-06, 'Rn': 0.5, 'SNR': 1e2}
    
    def set_status(self, status, channel=1):
        return

    def get_sweep_mode(self):
        return 0

    def get_sweep_channels(self):
        return (1, 2)
    
    def get_sweep_bias(self):
        return 0

    def set_func(self, func, *args, **kwargs):
        self.func, self.args, self.kwargs = func, args, kwargs

    def take_IV(self, sweep):
        start, stop, step, _ = sweep
        x = np.array([np.sign(val)*round(np.abs(val), -int(np.floor(np.log10(np.abs(step))))+1) for val in np.linspace(start, stop, int(round(np.abs(start-stop)/step+1)))])  # round to overcome missing precision of np.linspace
        y = self.func(x, *self.args, **self.kwargs)
        return x, y
    
    def get_parameters(self):
        return {}
    
    def get(self, param, **kwargs):
        return



# -*- coding: utf-8 -*-
"""
Remote interface service (RIS) for QKIT

The service forks a zerorpc RPC server in a separate thread. 
zerorpc manages incoming RPC requests concurrently on functions defined in QKIT_visible()
@author: HR@KIT 2018

#################################################################################
WARNING: Be careful in exposing new functions here:

The call may interrupt a running measurement or cause other race conditions.
Make sure the function is either GIL threadsave (atomic or builtin functions) or 
properly made threadsave by threading.Lock or similar.
#################################################################################

"""
import qkit
import threading
import zmq
import zerorpc
import logging

class flow(object):
    def stop_measure(self):
        return qkit.flow.stop()
    def hallo(self):
        return "hallo"

class QKIT_visible(object):
    """ Various convenience methods. """
    def __init__(self):
        self.flow = flow
    def __call__(self, func):
        return eval(func)

    def get_cfg(self):
        "returns the qkit.cfg dict"
        return qkit.cfg
    def get_current(self):
        pass
        #return qkit.last[-1]
    def get_last(self):
        pass
        #return qkit.last[-2]
    def stop_measure(self):
        pass
        #return qkit.flow.stop()
    def ex(self,func):
        if qkit.cfg.get("ris_debug",False):
            return eval(func)
        else:
            return False
    
    def get_instruments(self):
        return qkit.instruments.get_instrument_names()
    def get_instrument_param(selfs,instrument,parameter):
        return qkit.instruments.get(instrument).get(parameter)
    def get_all_instrument_params(self):
        return_dict = {}
        for ins in qkit.instruments.get_instruments().values():
            param_dict = {}
            for param in ins.get_parameter_names():
                param_dict.update({param: [ins.get(param, query=False), ins.get_parameter_options(param).get('units', "")]})
            return_dict.update({ins.get_name(): param_dict})
        return return_dict


class RISThread(object):
    def __init__(self):
        self.ris = ""
        logging.info(__name__+"starting RIS ...")
        thread = threading.Thread(target=self.run, name="ris", args=())
        thread.daemon = True                            # Daemonize thread
        thread.start()                                  # Start the execution
    def run(self):
        
        host = qkit.cfg.get("ris_host","127.0.0.1")
        port = int(qkit.cfg.get("ris_port",5700))
        socket_bound = False
        for newport in range(port,port+20):
            try:
                self.ris = zerorpc.Server(QKIT_visible())
                self.ris.bind("tcp://%s:%d"%(host,newport))
                host = qkit.cfg.get("ris_host","127.0.0.1")
                qkit.cfg["ris_host"] = host
                qkit.cfg["ris_port"] = newport
                socket_bound = True
                break
            except zmq.ZMQError as e:
                logging.debug(e)
                socket_bound = False
                self.ris.stop()
                del self.ris
        if not socket_bound:
            logging.warning("RIS: address/port in use: ZMQError. \nMaybe another 20 instances of QKIT are running?")
            logging.warning("Not starting RIS.")
            return False
        # run the thread
        self.ris.run()
    def stop(self):
            try:
                self.ris.stop()
            except:
                pass

if __name__ == "__main__":
    RIS = RISThread()



# -*- coding: utf-8 -*-
# Micha Wildermuth, micha.wildermuth@kit.edu 2020

from qkit.core.instrument_base import Instrument
import time


class quartz_dummy(Instrument)::
    '''
    This is a driver for a dummy quartz oscillator as used for sputter deposition monitoring qkit.services.qdepokit.sdi.

    Usage:
        Initialize with
        <name> = qkit.instruments.create('<name>', 'mfc_dummy')
    '''
    def __init__(self, name):
        self.__name__ = __name__
        Instrument.__init__(self, name, tags=['virtual'])
        self.t0 = None
        self.add_function('get_rate')
        self.add_function('get_thickness')


    def get_rate(self, nm=True):
        rate = 1.
        return rate

    def get_thickness(self, nm=True):
        if self.t0 is None:
            self.t0 = time.time()
        return self.get_rate()*(time.time()-self.t0)


import serial
import time
import sys
import zerorpc


class ERASynth(object):
    def __init__(self, ttyname, verb=True):
        s = '/dev/tty'
        self.path = s + str(ttyname)
        self.ser = serial.Serial(self.path)
        self.ser.baudrate = 115200
        self.verbose = verb

        if(self.ser.is_open == True):
            s = ""
            while s.find("HTTP server started") == -1:
                s += self.ser.readline().decode("ascii")  # .rstrip()
                if(self.verbose):
                    # sys.stdout.write(s)
                    print(s)
            print("Device: " + self.path)
            print("Initialization complete")
        else:
            print("Error: Establishing connection failed!")
            print("Check permissions and connection.")

    def __del__(self):
        self.ser.close()

    def readln(self):
        s = self.ser.readline().decode("ascii").rstrip()
        if(self.verbose):
            print(s)
            return s
        else:
            return s

    def enableout(self):
        self.ser.write(b'>P01\r\n')
        return self.readln()

    def disableout(self):
        self.ser.write(b'>P00\r\n')
        return self.readln()

    def setfrequency(self, f):
        self.ser.write(b'>F' + str(int(f)).encode() + b'\r\n')
        return self.readln()

    def setamplitude(self, a):
        self.ser.write(b'>A' + str(a).encode() + b'\r\n')
        return self.readln()

    def setrefint(self):
        self.ser.write(b'>P10\r\n')  # Clock Ref internal
        return self.readln()

    def setrefext(self):
        self.ser.write(b'>P11\r\n')  # Clock Ref internal
        return self.readln()

    def setrefocxo(self):
        self.ser.write(b'>P51\r\n')  # OCXO
        return self.readln()

    def setreftcxo(self):
        self.ser.write(b'>P50\r\n')  # TCXO
        return self.readln()

    def readall(self):
        self.ser.write(b'>RA\r\n')  # Readl all as json
        return self.readln()

    def readdiag(self):
        self.ser.write(b'>RD\r\n')
        return self.readln()


s = zerorpc.Server(ERASynth("ACM0", False))
s.bind("tcp://0.0.0.0:4242")
s.run()

'''
synth1 = ERASynth('ACM0')

synth1.setfrequency(6,9)
synth1.setamplitude(-22.2335)
synth1.enableout()
time.sleep(3)
synth1.disableout()
'''



# -*- coding: utf-8 -*-

"""
@author: hannes.rotzinger@kit.edu / 2015, 2016, 2017
         marco.pfirrmann@kit.edu / 2016, 2017
@license: GPL
"""


import sys
import qkit

# support both PyQt4 and 5
in_pyqt5 = False
in_pyqt4 = False
try:
    from PyQt5 import QtCore
    import PyQt5.QtWidgets as QtGui
    in_pyqt5 = True
except ImportError as e:
    pass

if not in_pyqt5:
    try:
        from PyQt4 import QtCore, QtGui
        in_pyqt4 = True
    except ImportError:
        print("import of PyQt5 and PyQt4 failed. Install one of those.")
        sys.exit(-1)


try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_MainWindow(object):
    """Ui_MainWindow class creates the frame of the  h5 file.

    The main window features some general interface buttons for file selection
    and live update as well as a interactive display to show and interact
    with the datasets in the populated dataset tree.
    """
    def setupUi(self, MainWindow):
        """setupUi creates the main window based on QMainWindow.

        Args:
            self: Object of the PlotWindow class.
            MainWindow: QMainWindow object.
        Returns:
            No return variable. The function operates on the given object.
        """
        MainWindow.setObjectName(_fromUtf8("MainWindow"))
        MainWindow.resize(271, 554)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.gridLayout_2 = QtGui.QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName(_fromUtf8("gridLayout_2"))
        self.verticalLayout = QtGui.QVBoxLayout()
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.gridLayout = QtGui.QGridLayout()
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 1, 1, 1)
        spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 0, 1, 1, 1)
        self.FileButton = QtGui.QPushButton(self.centralwidget)
        self.FileButton.setObjectName(_fromUtf8("FileButton"))
        self.gridLayout.addWidget(self.FileButton, 0, 0, 1, 1)
        self.liveCheckBox = QtGui.QCheckBox(self.centralwidget)
        self.liveCheckBox.setObjectName(_fromUtf8("liveCheckBox"))
        self.gridLayout.addWidget(self.liveCheckBox, 1, 2, 1, 1)
        self.updateButton = QtGui.QPushButton(self.centralwidget)
        self.updateButton.setObjectName(_fromUtf8("updateButton"))
        self.gridLayout.addWidget(self.updateButton, 1, 0, 1, 1)
        self.refreshTime = QtGui.QDoubleSpinBox(self.centralwidget)
        self.refreshTime.setMinimum(1.0)
        self.refreshTime.setMaximum(999.99)
        self.refreshTime.setProperty("value", 2.0)
        self.refreshTime.setObjectName(_fromUtf8("refreshTime"))
        self.gridLayout.addWidget(self.refreshTime, 1, 3, 1, 1)
        self.heardBeat = QtGui.QRadioButton(self.centralwidget)
        self.heardBeat.setText(_fromUtf8(""))
        self.heardBeat.setObjectName(_fromUtf8("heardBeat"))
        self.gridLayout.addWidget(self.heardBeat, 1, 4, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.treeWidget = QtGui.QTreeWidget(self.centralwidget)
        self.treeWidget.setObjectName(_fromUtf8("treeWidget"))
        self.treeWidget.headerItem().setText(0, _fromUtf8("1"))
        self.verticalLayout.addWidget(self.treeWidget)
        self.Dataset_properties = QtGui.QPlainTextEdit(self.centralwidget)
        self.Dataset_properties.setReadOnly(True)
        self.Dataset_properties.setObjectName(_fromUtf8("Dataset_properties"))
        self.verticalLayout.addWidget(self.Dataset_properties)
        self.gridLayout_2.addLayout(self.verticalLayout, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtGui.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 271, 22))
        self.menubar.setObjectName(_fromUtf8("menubar"))
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setObjectName(_fromUtf8("statusbar"))
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow", None))
        self.FileButton.setText(_translate("MainWindow", "File", None))
        self.liveCheckBox.setText(_translate("MainWindow", "live", None))
        self.updateButton.setText(_translate("MainWindow", "Update", None))


# -*- coding: utf-8 -*-
# timedomain.py analysis class for qkit time domain measurements data
# Micha Wildermuth, micha.wildermuth@kit.edu 2023

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
from qkit.analysis.qdata import qData
from qkit.analysis.qfit import QFIT

""" Error calculations with uncertainties package """
import uncertainties as uncert
from uncertainties import ufloat, unumpy as unp


class timedomain(qData):
    def __init__(self):
        super().__init__()
        self.qfit = QFIT()

    def load(self, uuid):
        super().load(uuid=uuid)
        if self.m_type != 'timedomain':
            raise AttributeError('No timedomain data loaded. Use data acquired with timedomain measurement class or general qData class.')
        self.scan_dim = self.df.data.amplitude_0.attrs['ds_type']  # scan dimension (1D, 2D, ...)
        self._get_xy_parameter(self.df.data.amplitude_0)

    def open_qviewkit(self, uuid=None, ds=None):
        """
        Opens qkit measurement data with given uuid <uuid> in qviewkit.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and opened in qviewkit.
        ds: str | list(str)
            Datasets that are opened instantaneously. Default for time-domain data is 'amplitude_0' and 'phase_0'.

        Returns
        -------
        None
        """
        if ds is None:
            ds = [_ds if _ds in self.df.data.__dict__.keys() else None for _ds in ['amplitude_0', 'phase_0']]
        super().open_qviewkit(uuid=uuid, ds=ds)

    def setup_T1_fit(self):
        self.qfit.load(self.path,
                       entries=['delay',
                                'phase_avg_0',
                                'amplitude_avg_0',
                                'phase_0',
                                'amplitude_0'])
        self.qfit.rotate_IQ_plane()

    def fit_T1(self, **kwargs):
        self.qfit.fit_exp(**kwargs)
        return ufloat(nominal_value=self.qfit.popt[0],
                      std_dev=np.sqrt(self.qfit.pcov[0, 0]))

    def setup_Ramsey_fit(self):
        self.qfit.load(self.path,
                       entries=['delay',
                                'phase_avg_0',
                                'amplitude_avg_0',
                                'phase_0',
                                'amplitude_0'])
        self.qfit.rotate_IQ_plane()

    def fit_Ramsey(self, **kwargs):
        self.qfit.fit_damped_sine(**kwargs)
        return ufloat(nominal_value=self.qfit.popt[1],
                      std_dev=np.sqrt(self.qfit.pcov[1, 1]))

    def setup_Echo_fit(self):
        self.qfit.load(self.path,
                       entries=['delay',
                                'phase_avg_0',
                                'amplitude_avg_0',
                                'phase_0',
                                'amplitude_0'])
        self.qfit.rotate_IQ_plane()

    def fit_Echo(self, **kwargs):
        self.qfit.fit_exp(**kwargs)
        return ufloat(nominal_value=self.qfit.popt[0],
                      std_dev=np.sqrt(self.qfit.pcov[0, 0]))




# filename: Virtual_VCA.py
# Tim Wolz <tim.wolz@.kit.edu>, 5/2017 updated 09/2018

# virtual VCA instrument to be used as a wrapper for several DACs
# so far the IVVI and DC_DAC are implemented

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
import types
import numpy as np
from scipy import interpolate

volts = np.linspace(0, 8, 17)
attenuation0408 = [0, 1.11278882e+00, 3.73766693e+00, 7.14756693e+00,
                   1.00470419e+01, 1.30124250e+01, 1.53641026e+01, 1.76756578e+01,
                   2.00411519e+01, 2.22864997e+01, 2.45740138e+01, 2.65537780e+01,
                   2.84989597e+01, 3.03834811e+01, 3.21356836e+01, 3.40573527e+01,
                   3.57549752e+01]  # These are measurement values, insertion loss not considered
# ToDo include measurment values of 0812 vca for interpolation
attenuation0812 = []


class Virtual_VCA(Instrument):
    """
    Wrapper for the VCA that allows to set the attenuation without knowing the underlying DAC mechanisms.
    """
    def __init__(self, name, dac_device, dac_port, freqrange):
        """
        inits the VCA (virtual instrument)
        :param name: qkit instrument name
        :param dac_device: either IVVI (must be set to positive values) or DC_DAC
        :param dac_port: chosen output port / channel of your DAC-device
        :param freqrange: frequency range of the vca as string either '0408' or '0812'
        """
        Instrument.__init__(self, name, tags=['virtual'])
        if freqrange is '0408':
            self.volts_interpol = interpolate.interp1d(attenuation0408, volts)
            self.att_interpol = interpolate.interp1d(volts, attenuation0408)

        # ToDo:
        # elif freqrange is '0812':
        #    self.volts_interpol=interpolate.interp1d(attenuation0812, volts)
        #    self.att_interpol=interpolate.interp1d(volts, attenuation0812)

        self.add_parameter('attenuation', type=float, flags=Instrument.FLAG_GETSET, units='dB', minval=0,
                           maxval=30)
        self._dac = _Dac(dac_device, dac_port)

    def do_set_attenuation(self, att):
        """
        sets the attenuation of the vca
        :param att in dB (float)
        """
        voltage = self.volts_interpol(att) / 2  # 2 bc of amplifier
        self._dac.set_dac_voltage(voltage)

    def do_get_attenuation(self):
        """
        returns the attenuation of the vca
        """
        voltage = self._dac.get_dac_voltage()
        return self.att_interpol(voltage*2)


class _Dac:
    """
    wrapper class so that Virtual_VCA works with several different Dacs.
    Please just append other devices here.
    """
    def __init__(self, dac_device, dac_port):
        self.dac_device = dac_device
        self.dac_port = dac_port

    def set_dac_voltage(self, voltage):
        if self.dac_device.get_type()[0:4] == 'IVVI':
            mv = voltage * 1000
            self.dac_device.set_dac(self.dac_port, mv, dacrange=(0, 4000))
        elif self.dac_device.get_type()[0:6] == 'DC_DAC':
            self.dac_device.do_set_voltage(voltage, self.dac_port)

    def get_dac_voltage(self):
        if self.dac_device.get_type()[0:4] == 'IVVI':
            voltage = self.dac_device.get_dac(self.dac_port, dacrange=(0, 4000)) / 1000
        elif self.dac_device.get_type()[0:6] == 'DC_DAC':
            voltage = self.dac_device.do_get_voltage(self.dac_port)
        return voltage
